/*
 * js-linkedmap.js v1.0.0
 * (c) 2018-2019 Jesse Feng
 * Released under the MIT License.
 */
'use strict';

var isArrayLike = require('celia/isArrayLike');
var _forEach = require('celia/_forEach');
var forOwn = require('celia/forOwn');
var append = require('celia/_append');
var alias = require('celia/alias');

var LinkedMap = function LinkedMap(iterator) {
  this
    .clear()
    .putAll(iterator);
};

LinkedMap.prototype._clone = function _clone (copy) {
  copy._link = this._link.slice(0);

  var copyMap = Object.create(null);
  forOwn(this._map, function (val, key) {
    copyMap[key] = val;
  });
  copy._map = copyMap;
  return copy;
};

/**
 * 自定义扩展获取 hash 值的方法
 * @param {any} key
 */
LinkedMap.prototype._getHash = function _getHash (key) {
  return String(key);
};

LinkedMap.prototype.clear = function clear () {
  this._link = [];
  this._map = Object.create(null);
  return this;
};

LinkedMap.prototype.clone = function clone () {
  return this._clone(new LinkedMap());
};

LinkedMap.prototype.containsKey = function containsKey (key) {
  return (key in this._map);
};

LinkedMap.prototype.containsValue = function containsValue (value) {
  var ref = this;
    var _link = ref._link;
    var _map = ref._map;
  return _link.some(function (key) { return _map[key] === value; });
};

LinkedMap.prototype.forEach = function forEach (callback) {
  var ref = this;
    var _link = ref._link;
    var _map = ref._map;
  _forEach(_link, function (key) {
    return callback(_map[key], key);
  });
  return this;
};

LinkedMap.prototype.get = function get (key, defaultValue) {
  key = this._getHash(key);
  var value = this._map[key];
  return value === undefined
    ? defaultValue
    : value;
};

LinkedMap.prototype.isEmpty = function isEmpty () {
  return this._link.length === 0;
};

LinkedMap.prototype.keys = function keys (callback) {
  return callback
    ? _forEach(this._link, callback)
    : this._link.slice(0);
};

LinkedMap.prototype.put = function put (key, value) {
  key = this._getHash(key);
  var ref = this;
    var _link = ref._link;
    var _map = ref._map;
  !(key in _map) && append(_link, key);
  _map[key] = value;
  return this;
};

LinkedMap.prototype.putAll = function putAll (iterator) {
    var this$1 = this;

  var i = 0;

  isArrayLike(iterator)
    ? _forEach(iterator, function (ref) {
        var key = ref[0];
        var value = ref[1];

      this$1.put(key, value);
      i++;
    })
    : forOwn(iterator, function (value, key) {
      this$1.put(key, value);
      i++;
    });
  return i;
};

LinkedMap.prototype.remove = function remove (key, specifiedValue) {
  var len = arguments.length;
  var result = false;
  var map = this._map;

  switch (len) {
    case 1:
      key = this._getHash(key);
      if (key in map) {
        delete map[key];
        result = true;
      }
      break;
    case 2:
      key = this._getHash(key);
      if ((key in map) && map[key] === specifiedValue) {
        delete map[key];
        result = true;
      }
      break;
    default:
      return false;
  }

  // 联动删除下标集合
  if (result) {
    var ref = this;
      var _link = ref._link;
    _link.splice(_link.indexOf(key), 1);
  }
  return result;
};

LinkedMap.prototype.replace = function replace (key, oldValue, newValue) {
  var len = arguments.length;
  var map = this._map;

  switch (len) {
    case 2:
      key = this._getHash(key);
      if ((key in map)) {
        map[key] = oldValue;
        return true;
      }
      break;
    case 3:
      key = this._getHash(key);
      if ((key in map) && map[key] === oldValue) {
        map[key] = newValue;
        return true;
      }
      break;
  }

  return false;
};

LinkedMap.prototype.size = function size () {
  return this._link.length;
};

LinkedMap.prototype.toString = function toString () {
  return '[object LinkedMap]';
};

LinkedMap.prototype.values = function values (callback) {
  var ref = this;
    var _link = ref._link;
    var _map = ref._map;
  return callback
    ? _forEach(_link, function (key, i) {
      return callback(_map[key], i);
    })
    : _link.map(function (key) { return _map[key]; });
};

var proto = LinkedMap.prototype;

if (Symbol && Symbol.iterator) {
  proto[Symbol.iterator] = function () {
    var ref = this;
    var _link = ref._link;
    var _map = ref._map;
    var i = 0;
    var len = _link.length;
    return {
      next: function next() {
        var key = _link[i];
        return { value: [key, _map[key], i], done: ++i > len };
      }
    };
  };
}

alias(proto, {
  keys: 'keySet',
  put: 'set',
  remove: 'delete'
});

module.exports = LinkedMap;
